import uuid
import asyncio
import websockets
import json
from typing import Dict, Any
import os
from ..mayalabs import authenticate
from ..consts import api_base_url, api_ws_url


class PacTask:
    @authenticate
    def __init__(self, type: str, opts: Dict[str, Any], api_key=None):
        self.type = type
        self.opts = opts
        self.api_key = api_key
        self.done_future = asyncio.Future()

    async def execute(self):
        connection_id = uuid.uuid4()
        url = f"{api_ws_url}?connId={connection_id}&apiKey={self.api_key}"
        
        async with websockets.connect(url) as socket:
            task_id = uuid.uuid4()

            message = {
                "task_id": str(task_id),
                "type": self.type,
                "data": self.opts,
            }
            await socket.send(json.dumps(message))

            async for message in socket:
                msg_object = json.loads(json.loads(message))
                if msg_object["task_id"] == str(task_id):
                    data = msg_object["data"]
                    if isinstance(data, str):
                        data = json.loads(data)
                    self.done_future.set_result(data)
                    break

    async def done(self):
        await self.done_future


class GenerateTask(PacTask):
    def __init__(self, session_id: str):
        super().__init__(
            type="GENERATE",
            opts={"session_id": session_id},
        )


class InstructTask(PacTask):
    def __init__(self, session_id: str, instruction: str, from_scratch: bool):
        super().__init__(
            type="INSTRUCT",
            opts={
                "session_id": session_id,
                "instruction": instruction,
                "from_scratch": from_scratch,
            },
        )




import random

adjectives = [
	'Adoring',
	'Amazing',
	'Awesome',
	'Beautiful',
	'Blissful',
	'Bold',
	'Brave',
	'Busy',
	'Charming',
	'Clever',
	'Compassionate',
	'Competent',
	'Condescending',
	'Confident',
	'Cool',
	'Cranky',
	'Crazy',
	'Dazzling',
	'Determined',
	'Distracted',
	'Dreamy',
	'Eager',
	'Ecstatic',
	'Elastic',
	'Elated',
	'Elegant',
	'Eloquent',
	'Epic',
	'Exciting',
	'Fervent',
	'Festive',
	'Flamboyant',
	'Focused',
	'Friendly',
	'Frosty',
	'Funny',
	'Gallant',
	'Gifted',
	'Goofy',
	'Gracious',
	'Great',
	'Happy',
	'Hardcore',
	'Heuristic',
	'Hopeful',
	'Hungry',
	'Infallible',
	'Inspiring',
	'Intelligent',
	'Interesting',
	'Jolly',
	'Jovial',
	'Keen',
	'Kind',
	'Laughing',
	'Lucid',
	'Magical',
	'Modest',
	'Musing',
	'Mystifying',
	'Nervous',
	'Nice',
	'Nifty',
	'Nostalgic',
	'Objective',
	'Optimistic',
	'Peaceful',
	'Pedantic',
	'Pensive',
	'Practical',
	'Priceless',
	'Quirky',
	'Quizzical',
	'Recursing',
	'Relaxed',
	'Reverent',
	'Serene',
	'Sharp',
	'Silly',
	'Sleepy',
	'Stoic',
	'Strange',
	'Stupefied',
	'Suspicious',
	'Sweet',
	'Trusting',
	'Unruffled',
	'Upbeat',
	'Vibrant',
	'Vigilant',
	'Vigorous',
	'Wizardly',
	'Wonderful',
	'Zealous',
	'Zen',
];

# thanks docker!
scientists = [
	'Agnesi',
	'Albattani',
	'Allen',
	'Almeida',
	'Antonelli',
	'Archimedes',
	'Aryabhata',
	'Austin',
	'Babbage',
	'Banach',
	'Banzai',
	'Bardeen',
	'Bartik',
	'Bassi',
	'Beaver',
	'Bell',
	'Benz',
	'Bhabha',
	'Bhaskara',
	'Black',
	'Blackburn',
	'Blackwell',
	'Bohr',
	'Bohm',
	'Booth',
	'Borg',
	'Bose',
	'Bouman',
	'Boyd',
	'Brahmagupta',
	'Brattain',
	'Brown',
	'Buck',
	'Burnell',
	'Cannon',
	'Carson',
	'Cartwright',
	'Carver',
	'Cerf',
	'Chandrasekhar',
	'Chatelet',
	'Chatterjee',
	'Chaum',
	'Clarke',
	'Cohen',
	'Colden',
	'Cori',
	'Cray',
	'Curie',
	'Curran',
	'Darwin',
	'Davinci',
	'Dewdney',
	'Dhawan',
	'Diffie',
	'Dirac',
	'Driscoll',
	'Dubinsky',
	'Easley',
	'Edison',
	'Einstein',
	'Elion',
	'Ellis',
	'Engelbart',
	'Euclid',
	'Euler',
	'Faraday',
	'Feistel',
	'Fermat',
	'Fermi',
	'Feynman',
	'Franklin',
	'Gagarin',
	'Galileo',
	'Galois',
	'Ganguly',
	'Gates',
	'Gauss',
	'Germain',
	'Goldberg',
	'Goldstine',
	'Goldwasser',
	'Golick',
	'Goodall',
	'Gould',
	'Haibt',
	'Hamilton',
	'Haslett',
	'Hawking',
	'Heisenberg',
	'Hellman',
	'Hermann',
	'Herschel',
	'Hertz',
	'Hofstadter',
	'Hoover',
	'Hopper',
	'Hugle',
	'Ishizaka',
	'Jackson',
	'Jang',
	'Jemison',
	'Jennings',
	'Jepsen',
	'Joliot',
	'Jones',
	'Kalam',
	'Kapitsa',
	'Kare',
	'Keldysh',
	'Keller',
	'Kepler',
	'Khayyam',
	'Khorana',
	'Kilby',
	'Kirch',
	'Knuth',
	'Lalande',
	'Lamarr',
	'Lamport',
	'Leakey',
	'Leavitt',
	'Lederberg',
	'Lehmann',
	'Lewin',
	'Lichterman',
	'Liskov',
	'Lovelace',
	'Lumiere',
	'Mahavira',
	'Margulis',
	'Matsumoto',
	'Maxwell',
	'Mayer',
	'Mccarthy',
	'Mcclintock',
	'Mclaren',
	'Mclean',
	'Mcnulty',
	'Meitner',
	'Mendel',
	'Mendeleev',
	'Meninsky',
	'Merkle',
	'Mestorf',
	'Mirzakhani',
	'Montalcini',
	'Moore',
	'Morse',
	'Moser',
	'Murdock',
	'Napier',
	'Nash',
	'Neumann',
	'Newton',
	'Nightingale',
	'Nobel',
	'Noether',
	'Northcutt',
	'Noyce',
	'Panini',
	'Pare',
	'Pascal',
	'Pasteur',
	'Payne',
	'Perlman',
	'Pike',
	'Poincare',
	'Poitras',
	'Ptolemy',
	'Raman',
	'Ramanujan',
	'Rhodes',
	'Ride',
	'Ritchie',
	'Robinson',
	'Roentgen',
	'Rosalind',
	'Rubin',
	'Saha',
	'Sammet',
	'Sanderson',
	'Satoshi',
	'Shamir',
	'Shannon',
	'Shaw',
	'Shirley',
	'Shockley',
	'Shtern',
	'Sinoussi',
	'Snyder',
	'Solomon',
	'Spence',
	'Stonebraker',
	'Sutherland',
	'Swanson',
	'Swartz',
	'Swirles',
	'Taussig',
	'Tesla',
	'Tharp',
	'Thompson',
	'Torvalds',
	'Turing',
	'Varahamihira',
	'Vaughan',
	'Villani',
	'Visvesvaraya',
	'Volhard',
	'Wescoff',
	'Wilbur',
	'Wiles',
	'Williams',
	'Williamson',
	'Wilson',
	'Wing',
	'Wozniak',
	'Wright',
	'Wu',
	'Yalow',
	'Yonath',
];

# write a function that returns a random scientist name, with a random adjective
def get_random_name():
    return random.choice(adjectives) + random.choice(scientists)


import asyncio

async def poll(fn=lambda: True, time_interval: int = 1000, timeout: int = 10000) -> None:
    async def poll_coroutine() -> None:
        while True:
            result = await fn()
            if result is True:
                break
            await asyncio.sleep(time_interval / 1000)

    try:
        await asyncio.wait_for(poll_coroutine(), timeout / 1000)
    except asyncio.TimeoutError as e:
        raise TimeoutError("Polling timed out") from e


from pydantic import BaseModel
from typing import List, Optional, Dict

class StepContainer(BaseModel):
    """
    A container for the step object. 
    """
    id : Optional[str]
    type : Optional[str]
    indent_lvl : Optional[int]
    from_node : Optional[str]
    lvl_idx : Optional[int] = 0
    prefix  : Optional[str] = ""
    parent_id : Optional[str] = ""
    error : Optional[str] = None
    children : Optional[List['StepContainer']] = []
    x_coord : Optional[int] = 0
    y_coord : Optional[int] = 0
    scrambled : Optional[bool] = False
    generated : Optional[bool] = False
    generation_id : Optional[str] = ""
    transformed : Optional[bool] = False
    deployed : Optional[bool] = False
    missing_reqs : Optional[List[dict]] = []
    use_cache : Optional[bool] = True

    def has_forks(self):
        assert_has_forks = False
        for child in self.children:
            if child.type == 'fork':
                assert_has_forks = True
        return assert_has_forks
    
    def dict(self):
        return {
            "id" : self.id,
            "step_obj" : self.step_obj.dict() if self.step_obj else None,
            "type" : self.type if self.type else "",
            "indent_lvl" : self.indent_lvl if self.indent_lvl else 0,
            "lvl_idx" : self.lvl_idx if self.lvl_idx else 0,
            "from_node" : self.from_node if self.from_node else "",
            "prefix" : self.prefix if self.prefix else "",
            "parent_id" : self.parent_id if self.parent_id else "",
            "generated" : self.generated,
            "scrambled" : self.scrambled,
            "transformed" : self.transformed,
            "generation_id" : self.generation_id,
            "error" : self.error,
            "deployed" : self.deployed,
            "children" : [child.dict() for child in self.children],
            "use_cache" : self.use_cache,
        }
    def parse_obj(self, json_raw):
        def create_step(step_raw):
            if not isinstance(step_raw, str):
                step = StepContainer()
                step.parse_obj(step_raw)
                return step


        data = json_raw
        if isinstance(data, dict):
            # print("IS DICT",data)
            self.id : str = data["id"]
            self.type : str = data["type"] if "type" in data.keys() else "linear"
            self.indent_lvl : int = data["indent_lvl"] if "indent_lvl" in data.keys() else 0
            self.lvl_idx : int = data["lvl_idx"] if "lvl_idx" in data.keys() else 0
            self.from_node : str = data["from_node"] if "node" in data.keys() else ""
            self.prefix  : str = data["prefix"] if "prefix" in data.keys() else ""
            self.parent_id : str = data["parent_id"] if "parent_id" in data.keys() else ""
            self.children : List['StepContainer'] = [create_step(child) for child in data["children"] if create_step(child)] if "children" in data.keys() else []
            self.generated : bool = data["generated"] if "generated" in data.keys() else False
            self.scrambled : bool = data["scrambled"] if "scrambled" in data.keys() else False
            self.transformed : bool = data["transformed"] if "transformed" in data.keys() else False
            self.error : bool = data["error"] if "error" in data.keys() else None
            self.deployed : bool = data["deployed"] if "deployed" in data.keys() else False
            self.generation_id : str = data["generation_id"] if "generation_id" in data.keys() else ""
            self.use_cache : bool = data["use_cache"] if "use_cache" in data.keys() else True

    
    def __str__(self):
        return self.to_string()
    
    def to_string(self):
        content = ""
        step_content = self.step_obj.step if self.step_obj.step else self.step_obj.sample_step
        content += step_content.replace("\n", "")
        if "(*)" not in content and "(!)" not in content:
            if not self.generated:
                ## add a (*) to the end of the printed step if ungenerated step
                content+=" (*)"
            if len(self.missing_reqs) > 0:
                ## add a (!) to the end of the printed step if requirements missing
                content+=" (!)"
            if self.error:
                if len(self.error) > 0:
                    content+=" (!)"
        else:
            if self.generated:
                content = content.replace("(*)", "")
            if not len(self.missing_reqs) > 0:
                ## add a (!) to the end of the printed step if requirements missing
                content = content.replace("(!)", "")
        return content

    def if_child_exists(self, child_id):
        def check_child_exists(step_children : List[StepContainer], id):
            check = False
            for child in step_children:
                print("JIPO:", child.id, id)
                if child.id == id:
                    check = True
                    break
                else:
                    check = check_child_exists(child.children, id)
            return check
        if self.id == child_id:
            return True
        else:
            return check_child_exists(self.children, child_id)




from .session import Session
from .worker import Worker
from .worker import WorkerClient
from .session import SessionClient
from .function import Function
from .mayalabs import auth, authenticate
from .consts import api_base_url, api_ws_url


import os
import click
from pathlib import Path

API_KEY_FILE = os.path.join(os.path.expanduser("~"), ".mayalabs")

def return_error(err):
    raise err

api_key = None

class Mayalabs():
    def __init__(self):
        self.api_key = None


auth = Mayalabs()

def authenticate(func):
    def wrapped_function(*args, **kwargs):
        ret = func(*args, api_key=auth.api_key, **kwargs)
        return ret
    return wrapped_function
# @click.group()
# @click.pass_context
# def cli(ctx):
#     ctx.obj = mayalabs()
#     ctx.obj.load_key()

# @cli.command()
# @click.option('--api_key', help='Your API key')
# @click.pass_obj
# def config(sdk, api_key):
#     sdk.set_key(api_key)
#     click.echo('API key configured.')


import asyncio
import websockets
import json
import sys
from colorama import init, Fore, Back, Style
from ..mayalabs import authenticate

init()

execution_events = [
    'debug',
    'status/#',
    'mayanodeerror',
    'mayanodewarning',
    'nodeexecstatus'
]

deploy_events = [
    'notification/runtime-deploy',
    'notification/#',
    'notification/node/#',
    # 'event-log/#'
]

# maya_log_prefix = Fore.MAGENTA + '[Maya]' + Style.RESET_ALL
maya_log_prefix = '[Maya]'

def rewrite_last_log_line(text):
    sys.stdout.write('\r')
    print(text)

def add_line_to_log(text):
    last_line = sys.stdout.readlines()[-1]
    sys.stdout.write('\r')


class WebsocketListener:
    def __init__(self, url):
        pass
        self.url = url
        self.handlers = {}
    
    def on(self, event, handler):
        if event not in self.handlers:
            self.handlers[event] = []
        self.handlers[event].append(handler)

    def handle_events(self, log_prefix, events):
        pass
        for event in events:
            if event['topic'] == 'nodeexecstatus':
                if (event['data']['status'] == 'running'):
                    nodeId = event['data']['nodeId']
                    print(log_prefix, Fore.CYAN + f'Running node: {nodeId}' + Style.RESET_ALL)
            
            elif event['topic'] == 'debug':
                msg_format = event['data']['format']
                content = event['data']['msg']

                level = event['data'].get('level', None)
                
                LOG_COLOR = Fore.YELLOW
                logLevel = 'debug'
                if level == 20:
                    LOG_COLOR = Fore.RED
                    logLevel = 'error'


                if msg_format == 'Object':
                    content = json.loads(content)
                nodeId = event['data']['id']
                print(log_prefix, LOG_COLOR + f'Received {logLevel} message from node {nodeId}: {content}' + Style.RESET_ALL)
            
            elif event['topic'] == 'notification/node/added':
                nodes = event['data']
                if len(nodes) == 0: return

                module_name = nodes[0]['module']
                module_version = nodes[0]['version']
                print(log_prefix, Fore.CYAN + f'Installed module {module_name}@{module_version}' + Style.RESET_ALL)
            
            # else:
            #     print('event received:', event)



                
    @authenticate
    async def start_listener(self, events=execution_events, log_prefix=maya_log_prefix, api_key=None):
        try:
            async with websockets.connect(self.url) as websocket:
                # print('Connected to worker at', self.url)
                self.websocket = websocket
                
                await websocket.send(json.dumps({ 'auth': api_key }))
                auth_response = await websocket.recv()
                # print('Auth response:', auth_response)

                for event in events:
                    # print('subscribing to', event)
                    await websocket.send(json.dumps({ 'subscribe': event }))
                while True:
                    message = await websocket.recv()
                    self.handle_events(log_prefix, json.loads(message))
                    # print('Received message from websocket:', message)
        except asyncio.CancelledError:
            await self.websocket.close()

    async def _disconnect(self, future):
        if self.websocket.open:
            await self.websocket.close()





import os
import requests
import asyncio
import aiohttp
import traceback
from .utils.poll import poll
from .utils.websocket import WebsocketListener
import time
from .consts import api_base_url, api_ws_url
from .mayalabs import authenticate
from colorama import init, Fore, Back, Style

# {'publishedSkillPacks': [], 'skillPacks': [], 'modules': [], 'externalModules': [], 'local': False, 'autoShutdownBehaviour': 'BY_LAST_USE', 'editorAutoDeploy': False, 'parent': None, 'alias': 'API_TEST', 'invalidNodes': [], 'createdAt': '2023-03-22T10:20:50.000Z', 'updatedAt': '2023-03-22T10:20:50.000Z', '_id': '641c89cb902176caaede0144', 'profileSlug': 'mayahq', 'name': 'TESTING API', 'status': 'STOPPED', 'deviceID': '5e0bbfe3717896af1cbb763b', 'deviceName': 'online-cpu', 'device': {'platform': 'cloud'}, 'thumbnail': 'https://maya-frontend-static.s3.ap-south-1.amazonaws.com/default.jpg', 'intents': [], 'url': 'https://rt-641c89cb902176caaede0144.mayahq.dev.mayalabs.io', 'deleted': False, 'createdBy': 'mayahq', 'updatedBy': 'mayahq', '__v': 0}

class Worker:
    def __init__(self) -> None:
        self.name : str = None
        self.alias : str = None
        self.id : str = None
        self.url : str = None
        self.status : str = None
        self.session_id : str = None
        self.ws_client : WebsocketListener = None

    def _init_from_api_response(self, response):
        self.name = response['name']
        self.alias = response['alias']
        self.id = response['_id']
        self.url = response['url']
        self.status = response['status']
        self.ws_client = WebsocketListener(
            url=self.url.replace('https', 'wss') + '/comms'
        )

    @staticmethod
    def get_by_id(id: str):
        worker_info = WorkerClient.get_worker(worker_id=id)
        worker = Worker()
        worker._init_from_api_response(worker_info['results'])
        return worker
        
    @staticmethod
    def get_by_alias(alias: str):
        return WorkerClient.get_worker_by_alias(alias)

    @staticmethod
    def search_by_name(name: str):
        return WorkerClient.search_worker_by_name(name=name)
    
    @staticmethod
    def create(name, alias):
        return WorkerClient.create_worker(worker_name=name, alias=alias)

    def start(self):
        return WorkerClient.start_worker(worker_id=self.id)

    def stop(self):
        return WorkerClient.stop_worker(worker_id=self.id)

    def delete(self):
        return WorkerClient.delete_worker(worker_id=self.id)

    def clear(self):
        pass

    def update(self):
        if self.id is not None:
            response = WorkerClient.get_worker(worker_id=self.id)
            self.parse_obj(response['results'])
        else:
            raise Exception("Worker ID is not set")
        
    @authenticate
    def attach_session(self, session_id, api_key=None):
        request = {
            'url': f"{api_base_url}/app/v2/brains/linkSessionToRuntime",
            'method': "post",
            'json': {
                'workspaceId': self.id,
                'sessionId': session_id
            },
            'headers': {
                'x-api-key': api_key,
            },
        }

        requests.request(**request)

    def call(self, msg : dict):
        if self.id is None:
            raise Exception("Worker ID is not set")
        if self.url is None:
            self.update()

        # loop = asyncio.get_event_loop()

        async def async_wrapper():
            call_task = asyncio.create_task(WorkerClient.call_worker(worker_url=self.url, msg=msg))
            log_task = asyncio.create_task(self.ws_client.start_listener())

        
            def stop_log_task(future):
                log_task.cancel()

            call_task.add_done_callback(stop_log_task)

            print(Style.BRIGHT + Fore.CYAN + '\nExecuting program on worker.\n' + Style.RESET_ALL)
            await asyncio.gather(call_task, log_task)

            return call_task, log_task

        call_task, log_task = asyncio.run(async_wrapper())
        return call_task.result()

    @classmethod
    def parse_obj(cls, obj):
        worker = cls()
        worker.id = obj.get('_id', None)
        worker.url = obj.get('url', None)
        worker.status = obj.get('status', None)
        worker.name = obj.get('name', None)
        worker.alias = obj.get('alias', None)
        worker.session_id = obj.get('sessionId', None)
        worker_ws_url = worker.url.replace('https', 'wss') + '/comms'
        worker.ws_client = WebsocketListener(url=worker_ws_url)
        return worker
    
    @classmethod
    def new(cls, name, alias=None):
        worker = WorkerClient.create_worker(worker_name=name, alias=alias)
        return worker



class WorkerClient:
    @staticmethod
    @authenticate
    def get_worker(worker_id, alias=None, api_key=None) -> Worker:
        if alias:
            request = {
                'url': f"{api_base_url}/app/v2/brains/{worker_id}",
                'method': "get",
                'json': {
                    'workspaceId': worker_id,
                    'alias': alias
                },
                'headers': {
                    'x-api-key': api_key,
                },
            }
        else:
            request = {
                'url': f"{api_base_url}/app/v2/brains/{worker_id}",
                'method': "get",
                'json': {
                    'workspaceId': worker_id,
                },
                'headers': {
                    'x-api-key': api_key,
                },
            }

        response = requests.request(**request)
        return response.json()
    
    @staticmethod
    @authenticate
    def get_worker_by_alias(alias, api_key=None) -> Worker:
        request = {
            'url': f"{api_base_url}/app/v2/brains/getByAlias/{alias}",
            'method': "get",
            'headers': {
                'x-api-key': api_key,
            },
        }

        response = requests.request(**request)
        results = response.json().get('results', None)
        if results is None:
            return None
        else:
            worker = Worker.parse_obj(results)
            return worker
    
    @staticmethod
    @authenticate
    def search_worker_by_name(self, name, api_key=None) -> Worker:
        request = {
            'url': f"{api_base_url}/app/v2/brains/search?name={name}",
            'method': 'get',
            'headers': {
                'x-api-key': api_key
            }
        }

        response = requests.request(**request)
        data = response.json()
        if len(data) > 0:
            worker = Worker.parse_obj(data[0]['results'])
            return worker
        else:
            return None

    @staticmethod
    @authenticate
    def create_worker(worker_name, alias, api_key=None) -> Worker:
        create_request = {
            'url': f"{api_base_url}/app/v2/brains",
            'method': "post",
            'json': {
                'name': worker_name,
                'alias': alias,
                'default': False,
                'device': {
                    'platform': 'cloud'
                }
            },
            'headers': {
                'x-api-key': api_key,
            },
        }
        response = requests.request(**create_request)
        worker = Worker.parse_obj(response.json()['results'])
        return worker
    
    @staticmethod
    @authenticate
    def start_worker(worker_id, auto_shutdown_behaviour=None, wait=False, api_key=None) -> Worker:
        start_request = {
            'url': f"{api_base_url}/app/v2/brains/start",
            'method': 'post',
            'json': {
                '_id': worker_id
            },
            'headers': {
                'x-api-key': api_key,
            }
        }

        if auto_shutdown_behaviour:
            start_request['json']['autoShutdownBehaviour'] = auto_shutdown_behaviour

        start_response = requests.request(**start_request)
        if wait:
            def start_confirmation_function():
                try:
                    health_state_response = requests.get(
                        f"{start_response.json()['results']['url']}/health?timesamp={int(time.time() * 1000)}",
                        timeout=2,
                        allow_redirects=True
                    )
                    return health_state_response.status_code == 200
                except requests.RequestException:
                    return False

            poll(start_confirmation_function, 1000, 120000)
        return start_response.json()
    
    @staticmethod
    @authenticate
    def stop_worker(worker_id, wait=False, api_key=None) -> Worker:
        stop_request = {
            'url': f"{api_base_url}/app/v2/brains/stop",
            'method': 'post',
            'json': {
                '_id': worker_id
            },
            'headers': {
                'x-api-key': api_key,
            }
        }

        stop_response = requests.request(**stop_request)
        # if wait:
        #     def start_confirmation_function():
        #         try:
        #             health_state_response = requests.get(
        #                 f"{stop_response.json()['results']['url']}/health?timesamp={int(time.time() * 1000)}",
        #                 timeout=2,
        #                 allow_redirects=True
        #             )
        #             return health_state_response.status_code == 200
        #         except requests.RequestException:
        #             return False

        #     poll(start_confirmation_function, 1000, 120000)
        return stop_response.json()

    @staticmethod
    @authenticate
    def delete_worker(worker_id, api_key=None):
        request = {
            'url': f"{api_base_url}/app/v2/brains/{worker_id}",
            'method': "delete",
            'json': {},
            'headers': {
                'x-api-key': api_key,
            },
        }

        response = requests.request(**request)
        return response.json

    @staticmethod
    @authenticate
    async def call_worker(worker_url, msg, api_key=None):
        async with aiohttp.ClientSession(headers={ 'x-api-key': api_key }) as session:
            async with session.post(f"{worker_url}/send-maya-message", json=msg) as response:
                response_json = await response.json()
                return response_json
    



from mayalabs import Session
from mayalabs import Worker, WorkerClient, SessionClient
from typing import Any, Dict
from .mayalabs import authenticate
import asyncio

class Function:
    # @authenticate
    def __init__(self, name, script, init = True):
        self.name : str = name
        self.script : str = script
        self.worker : Worker = None
        self.session : Session = None
        if init:
            self.init()

    @staticmethod
    def create(name, script):
        func = Function(
            name=name,
            script=script,
            init=False
        )

        worker = None
        try:
            worker = Worker.create(name=name, alias=name)
        except:
            raise Exception('Failed to create a worker for the function')

        session = None
        try:
            session = Session.new(script=script)
            worker.attach_session(session_id=session.id)
        except:
            raise Exception('Failed to create a new session for the worker')
        
        func.worker = worker
        func.session = session

        return func

    def init(self, api_key=None):
        """
        Initializes the function.
        Checks if the worker exists. If not, creates a new worker.
        Pass an api_key to override default and use a different profile.
        """
        if self.name is None:
            raise Exception("No name provided in the argument `name`")
        if self.script is None:
            raise Exception("No script provided in the argument `script`")
        try:
            # check if the worker exists    
            if api_key is None:
                self.worker = WorkerClient().get_worker_by_alias(self.name)
            else:
                self.worker = WorkerClient().get_worker_by_alias(self.name, api_key=api_key)

            if self.worker:
                if self.worker.session_id:
                    try:
                        self.session = Session.get(session_id=self.worker.session_id)
                    except Exception as e:
                        print("Session not found. Error:", e)
                        raise e
        except Exception as err:
            print("Worker not found.")
            raise err

    def deploy(self):
        """
        Deploys the function serverlessly.
        Creates a new session from script and deploys session to a worker.
        """
        if self.script is None:
            raise Exception("No script provided")

        if self.session is None:
            self.session = Session.new(script=self.script)

        if self.worker is None:
            self.worker = Worker.new(name=self.name, alias=self.name)

        self.session.deploy(worker_id=self.worker.id)

    def call(self, **kwargs : Dict) -> Dict:
        """
        Call the deployed function. 
        Arguments passed here are passed to the `msg` object in the script.
        A value like `msg['key']` can be accessed in the script by using {{key}}
        """
        return self.worker.call(msg = kwargs)
    
    def __call__(self, **kwds: Dict) -> Dict:
        """
        Call the deployed function. 
        Arguments passed here are passed to the `msg` object in the script.
        A value like `msg['key']` can be accessed in the script by using {{key}}
        """
        return self.call(**kwds)
    
    def clear(self):
        """
        Deletes the function by deleteing the associated session and the worker.
        """
        if self.session is not None:
            self.session.delete()
        if self.worker is not None:
            self.worker.delete()


import requests
import sys
import aiohttp

from .utils.pac_engine import GenerateTask, InstructTask
from .worker import WorkerClient, Worker
from .utils.name_gen import get_random_name
from .utils.websocket import WebsocketListener, deploy_events
import asyncio
import time
from time import sleep
import concurrent.futures
from .consts import api_base_url, api_ws_url
from .mayalabs import authenticate
from colorama import init, Fore, Back, Style


class SessionClient:

    @staticmethod
    @authenticate
    def get_session(session_id, alias=None, api_key=None):
        """get the session object by providing session identifier

        Args:
            session_id (_type_): _description_
            alias (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """
        request = {
            'url': f"{api_base_url}/pac/v1/session/{session_id}",
            'method': "get",
            'json': {},
            'headers': {
                'x-api-key': api_key,
            }
        }
        response = requests.request(**request)
        return response.json()
    
    @staticmethod
    @authenticate
    def create_session(from_script="", api_key=None) -> 'Session':
        request = {
            'url': f"{api_base_url}/pac/v1/session/new",
            'method': "post",
            'json': {
                'from_recipe': from_script
            },
            'headers': {
                'x-api-key': api_key,
            }
        }
        response = requests.request(**request)
        return response.json()
    
    @staticmethod
    @authenticate
    def list_sessions(api_key=None):
        request = {
            'url': f"{api_base_url}/pac/v1/sessions",
            'method': "get",
            'json': {},
            'headers': {
                'x-api-key': api_key,
            }
        }
        response = requests.request(**request)
        return response.json()
    
    @staticmethod
    @authenticate
    def delete_session(session_id, api_key=None):
        request = {
            'url': f"{api_base_url}/pac/v1/session/{session_id}",
            'method': "delete",
            'json': {},
            'headers': {
                'x-api-key': api_key,
            }
        }
        response = requests.request(**request)
        return response.json()
    
    # @authenticate
    # def deploy_session(session_id, workspace_id, api_key=None):
    #     request = {
    #         'url': f"{api_base_url}/pac/v1/session/deploy",
    #         'method': "post",
    #         'json': {
    #             'session_id': session_id,
    #             'workspace_id' : workspace_id,
    #         },
    #         'headers': {
    #             'x-api-key': api_key,
    #         }
    #     }
    #     response = requests.request(**request)
    #     try:
    #         return response.json()
    #     except:
    #         print('deploy error')

    @authenticate
    async def deploy_session(session_id, workspace_id, api_key=None):
        data = {
            'session_id': session_id,
            'workspace_id' : workspace_id,
        }
        async with aiohttp.ClientSession(headers={'x-api-key': api_key}) as session:
            async with session.post(f'{api_base_url}/pac/v1/session/deploy', json=data) as response:
                response_json = await response.json()
                return response_json
    
    # def session_parse(obj):
    #     # requests = {
    #     #     'url': f"{self.pac_base_url}/pac/v1/session/parse",
    #     #     'method': "post",
    #     #     'json': {
                
    #     #     },
    #     #     'headers': {
    #     #         'x-api-key': self.api_key,
            
    #     # }
    #     pass

class Session():
    def __init__(self, engine="pac-1"):
        self.id = None
        self.engine = engine
        self.script = ""
        self.worker : Worker = None

    @classmethod
    def new(cls, script=None):
        response = SessionClient.create_session(from_script=script)
        session = cls()
        session.parse_obj(response['response'])
        return session
   
    @classmethod
    def get(cls, session_id=None):
        response = SessionClient.get_session(session_id=session_id)
        session = cls()
        session.parse_obj(response['response'])
        return session
    
    def instruct(self, prompt, from_scratch=True):
        # Implement this method
        task = InstructTask(self.id, prompt, from_scratch=from_scratch)
        task.execute(prompt)
        pass

    def generate(self):
        task = GenerateTask(self.id)
        asyncio.create_task(task.execute())
        pass

    async def generate_async(self):
        task = GenerateTask(self.id)
        await task.execute()
        pass

    def check_worker_start(self):
        status = 0
        i = 0
        if (self.worker.status and self.worker.status != 'STARTED'):
            status = Fore.RED + 'PENDING' + Style.RESET_ALL
            print('[Maya]', "Worker status:", status)

        while self.worker.status and self.worker.status != "STARTED":
            i += 1
            # self.worker.update()
            worker_response = WorkerClient.get_worker(self.worker.id)
            if worker_response['results']:
                self.worker = Worker().parse_obj(worker_response['results'])
            time.sleep(2)

        started_status = Fore.GREEN + 'STARTED' + Style.RESET_ALL
        print('[Maya]', 'Worker status:', started_status)
        return

    def deploy(self, worker_id=None):
        # Implement this method
        # print(worker_id)
        def run_asyncio_coroutine(coroutine):
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            result = loop.run_until_complete(coroutine)
            loop.close()
            return result
        
        if worker_id is not None:
            try:
                self.worker = Worker.get_by_id(worker_id)
                print('[Maya]', "Found worker: ", self.worker.name)
            except:
                raise Exception("Worker not found")
        elif self.worker is None:
            print("No worker found, creating new worker...")
            random_name = "SDK:" + get_random_name()
            self.worker = WorkerClient.create_worker(worker_name=random_name, alias=random_name)
        else:
            raise Exception("Error: Could not find worker")
        
        if self.worker:
            if self.worker.status != "STARTED":
                print('[Maya]', "Starting worker: ", self.worker.name)
                self.worker.start()
            print(f'[{self.worker.name}]', Style.BRIGHT + Fore.CYAN + 'Generating program.' + Style.RESET_ALL)
            with concurrent.futures.ThreadPoolExecutor() as exec:
                future_1 = exec.submit(run_asyncio_coroutine, self.generate_async())
                result_2 = exec.submit(self.check_worker_start)
                result_1 = future_1.result()
                print(f'[{self.worker.name}]', Style.BRIGHT + Fore.GREEN + 'Generation successful.' + Style.RESET_ALL)
                result_2.result()
                # report all tasks done
            
            # loop = asyncio.get_event_loop()
            async def async_wrapper():
                deploy_task = asyncio.create_task(SessionClient.deploy_session(self.id, self.worker.id))
                log_task = asyncio.create_task(self.worker.ws_client.start_listener(events=deploy_events, log_prefix=f'[{self.worker.name}]'))
                def stop_log_task(future):
                    log_task.cancel()

                deploy_task.add_done_callback(stop_log_task)
                print(f'[{self.worker.name}]', Style.BRIGHT + Fore.CYAN + 'Deploying session to worker. Setting up dependencies.' + Style.RESET_ALL)
                await asyncio.gather(deploy_task, log_task)
                print(f'[{self.worker.name}]', Style.BRIGHT + Fore.GREEN + 'Deploy successful.' + Style.RESET_ALL)

                return deploy_task.result()
            asyncio.run(async_wrapper())
            
        else:
            raise Exception("Error: Could not find worker") 

    def delete(self):
        # Implement this method
        response = SessionClient.delete_session(self.id)
        return print(response)

    def parse_obj(self, obj):
        self.id = obj['session_id']
        self.script = obj['recipe']

    def update(self):
        response = SessionClient.get_session(self.id)
        print(response)
        self.parse_obj(response['response'])

    def to_string(self):
        pass

    def call(self, msg):
        call_task = asyncio.create_task(self.worker.call(msg))
        log_task = asyncio.create_task(self.worker.ws_client.create_listen_task())
        call_task.add_done_callback(log_task.cancel)

        tasks = asyncio.gather([call_task, log_task])
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(tasks)

        return response

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

